@@ -23,9 +23,11 @@
 #include "package.h"
 
 #include <qtemporarydir.h>
+#include <QResource>
 
 #include <karchive.h>
-#include <QDebug>
+#include "kpackage_debug.h"
+#include <QUrl>
 #include <ktar.h>
 #include <kzip.h>
 #include <KLocalizedString>
@@ -40,6 +42,7 @@
 #include "private/package_p.h"
 //#include "private/packages_p.h"
 #include "private/packagejob_p.h"
+#include "private/packageloader_p.h"
 
 namespace KPackage
 {
@@ -118,7 +121,7 @@ bool Package::isValid() const
 
         const QString foundPath = filePath(it.key(), {});
         if (foundPath.isEmpty()) {
-            //qWarning() << "Could not find required" << (it.value().directory ? "directory" : "file") << it.key() << "for package" << path() << "should be" << it.value().paths;
+            //qCWarning(KPACKAGE_LOG) << "Could not find required" << (it.value().directory ? "directory" : "file") << it.key() << "for package" << path() << "should be" << it.value().paths;
             d->valid = false;
             break;
         }
@@ -209,9 +212,10 @@ void Package::setAllowExternalPaths(bool allow)
 
 KPluginMetaData Package::metadata() const
 {
-    //qDebug() << "metadata: " << d->path << filePath("metadata");
+    //qCDebug(KPACKAGE_LOG) << "metadata: " << d->path << filePath("metadata");
     if (!d->metadata && !d->path.isEmpty()) {
         const QString metadataPath = filePath("metadata");
+
         if (!metadataPath.isEmpty()) {
             d->createPackageMetadata(metadataPath);
         } else {
@@ -255,7 +259,7 @@ QString PackagePrivate::unpack(const QString &filePath)
                mimeType.inherits(QStringLiteral("application/x-lzma"))) {
         archive = new KTar(filePath);
     } else {
-        //qWarning() << "Could not open package file, unsupported archive format:" << filePath << mimeType.name();
+        //qCWarning(KPACKAGE_LOG) << "Could not open package file, unsupported archive format:" << filePath << mimeType.name();
     }
     QString tempRoot;
     if (archive && archive->open(QIODevice::ReadOnly)) {
@@ -278,7 +282,7 @@ QString PackagePrivate::unpack(const QString &filePath)
 
         createPackageMetadata(tempRoot);
     } else {
-        //qWarning() << "Could not open package file:" << path;
+        //qCWarning(KPACKAGE_LOG) << "Could not open package file:" << path;
     }
 
     delete archive;
@@ -313,7 +317,7 @@ bool PackagePrivate::isInsidePackageDir(const QString &canonicalPath) const
             return true;
         }
     }
-    qWarning() << "Path traversal attempt detected:" << canonicalPath << "is not inside" << path;
+    qCWarning(KPACKAGE_LOG) << "Path traversal attempt detected:" << canonicalPath << "is not inside" << path;
     return false;
 }
 
@@ -323,7 +327,7 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
     if (!d->valid) {
         QString result = d->fallbackFilePath(fileType, filename);
         if (result.isEmpty()) {
-            // qDebug() << fileType << "file with name" << filename
+            // qCDebug(KPACKAGE_LOG) << fileType << "file with name" << filename
             //    << "was not found in package with path" << d->path;
         }
         return result;
@@ -340,14 +344,14 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
     if (fileType.size()) {
         const auto contents = d->contents.constFind(fileType);
         if (contents == d->contents.constEnd()) {
-            //qDebug() << "package does not contain" << fileType << filename;
+            //qCDebug(KPACKAGE_LOG) << "package does not contain" << fileType << filename;
             return d->fallbackFilePath(fileType, filename);
         }
 
         paths = contents->paths;
 
         if (paths.isEmpty()) {
-            //qDebug() << "no matching path came of it, while looking for" << fileType << filename;
+            //qCDebug(KPACKAGE_LOG) << "no matching path came of it, while looking for" << fileType << filename;
             d->discoveries.insert(discoveryKey, QString());
             return d->fallbackFilePath(fileType, filename);
         }
@@ -357,7 +361,7 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
     }
 
     //Nested loop, but in the medium case resolves to just one iteration
-//     qDebug() << "prefixes:" << d->contentsPrefixPaths.count() << d->contentsPrefixPaths;
+//     qCDebug(KPACKAGE_LOG) << "prefixes:" << d->contentsPrefixPaths.count() << d->contentsPrefixPaths;
     foreach (const QString &contentsPrefix, d->contentsPrefixPaths) {
         QString prefix;
         //We are an installed package
@@ -378,7 +382,7 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
             QFileInfo fi(file);
             if (fi.exists()) {
                 if (d->externalPaths) {
-                    //qDebug() << "found" << file;
+                    //qCDebug(KPACKAGE_LOG) << "found" << file;
                     d->discoveries.insert(discoveryKey, file);
                     return file;
                 }
@@ -386,7 +390,7 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
                 // ensure that we don't return files outside of our base path
                 // due to symlink or ../ games
                 if (d->isInsidePackageDir(fi.canonicalFilePath())) {
-                    //qDebug() << "found" << file;
+                    //qCDebug(KPACKAGE_LOG) << "found" << file;
                     d->discoveries.insert(discoveryKey, file);
                     return file;
                 }
@@ -394,10 +398,21 @@ QString Package::filePath(const QByteArray &fileType, const QString &filename) c
         }
     }
 
-    //qDebug() << fileType << filename << "does not exist in" << prefixes << "at root" << d->path;
+    //qCDebug(KPACKAGE_LOG) << fileType << filename << "does not exist in" << prefixes << "at root" << d->path;
     return d->fallbackFilePath(fileType, filename);
 }
 
+QUrl Package::fileUrl(const QByteArray &fileType, const QString &filename) const
+{
+    QString path = filePath(fileType, filename);
+    //construct a qrc:/ url or a file:/ url, the only two protocols supported
+    if (path.startsWith(QStringLiteral(":"))) {
+        return QUrl(QStringLiteral("qrc") + path);
+    } else {
+        return QUrl::fromLocalFile(path);
+    }
+}
+
 QStringList Package::entryList(const QByteArray &key) const
 {
     if (!d->valid) {
@@ -406,18 +421,18 @@ QStringList Package::entryList(const QByteArray &key) const
 
     QHash<QByteArray, ContentStructure>::const_iterator it = d->contents.constFind(key);
     if (it == d->contents.constEnd()) {
-        //qDebug() << "couldn't find" << key;
+        //qCDebug(KPACKAGE_LOG) << "couldn't find" << key;
         return QStringList();
     }
 
-    //qDebug() << "going to list" << key;
+    //qCDebug(KPACKAGE_LOG) << "going to list" << key;
     QStringList list;
     foreach (const QString &prefix, d->contentsPrefixPaths) {
-        //qDebug() << "     looking in" << prefix;
+        //qCDebug(KPACKAGE_LOG) << "     looking in" << prefix;
         foreach (const QString &path, it.value().paths) {
-            //qDebug() << "         looking in" << path;
+            //qCDebug(KPACKAGE_LOG) << "         looking in" << path;
             if (it.value().directory) {
-                //qDebug() << "it's a directory, so trying out" << d->path + prefix + path;
+                //qCDebug(KPACKAGE_LOG) << "it's a directory, so trying out" << d->path + prefix + path;
                 QDir dir(d->path + prefix + path);
 
                 if (d->externalPaths) {
@@ -432,7 +447,7 @@ QStringList Package::entryList(const QByteArray &key) const
                 }
             } else {
                 const QString fullPath = d->path + prefix + path;
-                //qDebug() << "it's a file at" << fullPath << QFile::exists(fullPath);
+                //qCDebug(KPACKAGE_LOG) << "it's a file at" << fullPath << QFile::exists(fullPath);
                 if (!QFile::exists(fullPath)) {
                     continue;
                 }
@@ -443,7 +458,7 @@ QStringList Package::entryList(const QByteArray &key) const
                     QDir dir(fullPath);
                     QString canonicalized = dir.canonicalPath() + QDir::separator();
 
-                    //qDebug() << "testing that" << canonicalized << "is in" << d->path;
+                    //qCDebug(KPACKAGE_LOG) << "testing that" << canonicalized << "is in" << d->path;
                     if (canonicalized.startsWith(d->path)) {
                         list += fullPath;
                     }
@@ -498,6 +513,7 @@ void Package::setPath(const QString &path)
         }
 
         if (QDir::isRelativePath(p)) {
+            //FIXME: can searching of the qrc be better?
             paths << QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, p, QStandardPaths::LocateDirectory);
         } else {
             const QDir dir(p);
@@ -506,7 +522,7 @@ void Package::setPath(const QString &path)
             }
         }
 
-        //qDebug() << "paths:" << p << paths << d->defaultPackageRoot;
+        //qCDebug(KPACKAGE_LOG) << "paths:" << p << paths << d->defaultPackageRoot;
     } else {
         const QDir dir(path);
         if (QFile::exists(dir.canonicalPath())) {
@@ -526,7 +542,16 @@ void Package::setPath(const QString &path)
     foreach (const QString &p, paths) {
         d->checkedValid = false;
         QDir dir(p);
+
         Q_ASSERT(QFile::exists(dir.canonicalPath()));
+
+        //if it has a contents.rcc, give priority to it
+        if (dir.exists(QStringLiteral("contents.rcc"))) {
+            d->rccPath = dir.absoluteFilePath(QStringLiteral("contents.rcc"));
+            QResource::registerResource(d->rccPath);
+            dir = QDir(QStringLiteral(":/") + defaultPackageRoot() + path);
+        }
+
         d->path = dir.canonicalPath();
          // canonicalPath() does not include a trailing slash (unless it is the root dir)
         if (!d->path.endsWith(QLatin1Char('/'))) {
@@ -607,7 +632,7 @@ QString Package::contentsHash() const
 QByteArray Package::cryptographicHash(QCryptographicHash::Algorithm algorithm) const
 {
     if (!d->valid) {
-        qWarning() << "can not create hash due to Package being invalid";
+        qCWarning(KPACKAGE_LOG) << "can not create hash due to Package being invalid";
         return QByteArray();
     }
 
@@ -622,10 +647,10 @@ QByteArray Package::cryptographicHash(QCryptographicHash::Algorithm algorithm) c
                 hash.addData(f.read(1024));
             }
         } else {
-            qWarning() << "could not add" << f.fileName() << "to the hash; file could not be opened for reading.";
+            qCWarning(KPACKAGE_LOG) << "could not add" << f.fileName() << "to the hash; file could not be opened for reading.";
         }
     } else {
-        qWarning() << "no metadata at" << metadataPath;
+        qCWarning(KPACKAGE_LOG) << "no metadata at" << metadataPath;
     }
 
     foreach (const QString &prefix, d->contentsPrefixPaths) {
@@ -792,6 +817,7 @@ KJob *Package::install(const QString &sourcePackage, const QString &packageRoot)
 {
     const QString src = sourcePackage;
     QString dest = packageRoot.isEmpty() ? defaultPackageRoot() : packageRoot;
+    KPackage::PackageLoader::self()->d->maxCacheAge = -1;
 
     //use absolute paths if passed, otherwise go under share
     if (!QDir::isAbsolutePath(dest)) {
@@ -801,8 +827,8 @@ KJob *Package::install(const QString &sourcePackage, const QString &packageRoot)
     if (!d->structure) {
         return nullptr;
     }
-    //qDebug() << "Source: " << src;
-    //qDebug() << "PackageRoot: " << dest;
+    //qCDebug(KPACKAGE_LOG) << "Source: " << src;
+    //qCDebug(KPACKAGE_LOG) << "PackageRoot: " << dest;
     KJob *j = d->structure.data()->install(this, src, dest);
     return j;
 }
@@ -811,6 +837,7 @@ KJob *Package::update(const QString &sourcePackage, const QString &packageRoot)
 {
     const QString src = sourcePackage;
     QString dest = packageRoot.isEmpty() ? defaultPackageRoot() : packageRoot;
+     KPackage::PackageLoader::self()->d->maxCacheAge = -1;
 
     //use absolute paths if passed, otherwise go under share
     if (!QDir::isAbsolutePath(dest)) {
@@ -820,14 +847,15 @@ KJob *Package::update(const QString &sourcePackage, const QString &packageRoot)
     if (!d->structure) {
         return nullptr;
     }
-    //qDebug() << "Source: " << src;
-    //qDebug() << "PackageRoot: " << dest;
+    //qCDebug(KPACKAGE_LOG) << "Source: " << src;
+    //qCDebug(KPACKAGE_LOG) << "PackageRoot: " << dest;
     KJob *j = d->structure.data()->update(this, src, dest);
     return j;
 }
 
 KJob *Package::uninstall(const QString &packageName, const QString &packageRoot)
 {
+     KPackage::PackageLoader::self()->d->maxCacheAge = -1;
     d->createPackageMetadata(packageRoot + QLatin1Char('/') + packageName);
     if (!d->structure) {
         return nullptr;
@@ -855,6 +883,13 @@ PackagePrivate::PackagePrivate(const PackagePrivate &other)
 
 PackagePrivate::~PackagePrivate()
 {
+    if (!rccPath.isEmpty()) {
+        //qresource register/unregisterpath is refcounted if we call it two times
+        //on the same path, the resource will actually be unregistered only when
+        //unregister is called 2 times
+        QResource::unregisterResource(rccPath);
+    }
+
     if (!tempRoot.isEmpty()) {
         QDir dir(tempRoot);
         dir.removeRecursively();
@@ -917,7 +952,7 @@ void PackagePrivate::updateHash(const QString &basePath, const QString &subPath,
                     hash.addData(f.read(1024));
                 }
             } else {
-                qWarning() << "could not add" << f.fileName() << "to the hash; file could not be opened for reading. "
+                qCWarning(KPACKAGE_LOG) << "could not add" << f.fileName() << "to the hash; file could not be opened for reading. "
                            << "permissions fail?" << info.permissions() << info.isFile();
             }
         }
@@ -951,7 +986,7 @@ void PackagePrivate::createPackageMetadata(const QString &path)
         metadata = new KPluginMetaData(md);
     } else {
         if (isDir) {
-            qWarning() << "No metadata file in the package, expected it at:" << path;
+            qCWarning(KPACKAGE_LOG) << "No metadata file in the package, expected it at:" << path;
         } else if (path.endsWith(QLatin1String(".desktop"))) {
             auto md = KPluginMetaData::fromDesktopFile(path, {QStringLiteral(":/kservicetypes5/kpackage-generic.desktop")});
             metadata = new KPluginMetaData(md);
@@ -986,7 +1021,7 @@ bool PackagePrivate::hasCycle(const KPackage::Package &package)
         //consider two packages the same if they have the same metadata
         if ((fastPackage->d->fallbackPackage->metadata().isValid() && fastPackage->d->fallbackPackage->metadata() == slowPackage->metadata()) ||
             (fastPackage->d->fallbackPackage->d->fallbackPackage && fastPackage->d->fallbackPackage->d->fallbackPackage->metadata().isValid() && fastPackage->d->fallbackPackage->d->fallbackPackage->metadata() == slowPackage->metadata())) {
-            qWarning() << "Warning: the fallback chain of " << package.metadata().pluginId() << "contains a cyclical dependency.";
+            qCWarning(KPACKAGE_LOG) << "Warning: the fallback chain of " << package.metadata().pluginId() << "contains a cyclical dependency.";
             return true;
         }
         fastPackage = fastPackage->d->fallbackPackage->d->fallbackPackage;
